<?php

declare(strict_types=1);

namespace SimAVRPHP\Opcode;

use SimAVRPHP\AVR;
use SimAVRPHP\Exceptions\UnsupportedInstructionException;
use Exception;

class Instruction
{
    /**
     * @var class-string[]
     */
    protected static $AVRInstructions = [
        \SimAVRPHP\Opcode\BRNE::class,
        \SimAVRPHP\Opcode\CALL::class,
        \SimAVRPHP\Opcode\CLI::class,
        \SimAVRPHP\Opcode\EOR::class,
        \SimAVRPHP\Opcode\JMP::class,
        \SimAVRPHP\Opcode\LDI::class,
        \SimAVRPHP\Opcode\NOP::class,
        \SimAVRPHP\Opcode\OUT::class,
        \SimAVRPHP\Opcode\RJMP::class,
        \SimAVRPHP\Opcode\SBCI::class,
        \SimAVRPHP\Opcode\SUBI::class,
    ];

    /**
     * @param array<string,int> $operands
     */
    protected function __construct(protected AVR $mcu, protected array $operands)
    {
    }

    /**
     * @return bool|array<string,int>
     */
    private static function decode(string $opcode, string $pattern): array|bool
    {
        $data = [];
        $pattern_length = strlen($pattern);
        if ($pattern_length != 16 && $pattern_length != 32) {
            throw new Exception('Pattern length invalid ' . $pattern_length);
        }

        for ($c = 0; $c < $pattern_length; $c++) {
            $token = $pattern[$c];
            $instructionBit = $opcode[$c];
            if ($token === "1" || $token === "0") {
                if ($instructionBit !== $token) {
                    return FALSE;
                }
            } else {
                if (! isset($data[$token])) {
                    $data[$token] = "";
                }

                $data[$token] .= $instructionBit;
            }
        }

        return array_map(
            fn(string $element): int => (int)bindec($element),
            $data
        );
    }

    public function execute(): void
    {
        if (is_callable([$this,'do'])) {
            call_user_func_array([$this,'do'], $this->operands);
        }
    }

    public static function createInstructionFromOpcode(AVR $mcu, string $opcode): Instruction
    {
        foreach (self::$AVRInstructions as $i) {
            $data = self::decode($opcode, constant($i . "::OPCODE"));
            if ($data !== FALSE) {
                $obj = new $i($mcu, $data);
                if ($obj instanceof Instruction) {
                    return $obj;
                }
            }
        }

        throw new UnsupportedInstructionException("Unsupported instruction", (int)bindec($opcode));
    }

    /**
     * @param array<string,int> $param
     */
    public function debug(string $string, array $param = []): void
    {
        $this->mcu->logger->debug($string, $param);
    }

    protected function getRegister(int $reg): int
    {
        return $this->mcu->registers->get($reg);
    }

    public static function signed(int $v, int $min, int $max): int
    {
        if ($v > $max + abs($min)) {
            throw new Exception("Not valid within range");
        }

        if ($v > $max) {
            return $min - 1 + $v - $max;
        }

        return $v;
    }
}
